// VideoMemoryAllocator class implementation
#include "pch.h"

#include "VideoMemoryAllocator.h"
#include "util.h"

// 1 GiB
#define HEAP_SIZE_LARGE 1024ull * 1024 * 1024
// 50 MiB
#define DEFAULT_ALLOCATION_AMOUNT 50ull * 1024 * 1024

namespace nf::client::render {
    VideoMemoryAllocator::VideoMemoryAllocator(const VkDevice& device, const VkPhysicalDevice& physicalDevice)
        : m_device(device)
        , m_physicalDevice(physicalDevice)
        , m_blocks()
    {}

    void VideoMemoryAllocator::allocateForBuffer(VkBuffer buffer, VkMemoryPropertyFlags memoryPropertyFlags, VideoMemoryAllocation& outAllocation) {
        VkMemoryRequirements memoryRequirements = {};
        vkGetBufferMemoryRequirements(m_device, buffer, &memoryRequirements);

        allocate(memoryRequirements, memoryPropertyFlags, outAllocation);
        vkBindBufferMemory(m_device, buffer, outAllocation.memory, outAllocation.offset);
    }

    void VideoMemoryAllocator::allocateForImage(VkImage image, VkMemoryPropertyFlags memoryPropertyFlags, VideoMemoryAllocation& outAllocation) {
        VkMemoryRequirements memoryRequirements = {};
        vkGetImageMemoryRequirements(m_device, image, &memoryRequirements);

        allocate(memoryRequirements, memoryPropertyFlags, outAllocation);
        vkBindImageMemory(m_device, image, outAllocation.memory, outAllocation.offset);
    }

    void VideoMemoryAllocator::deallocate(const VideoMemoryAllocation& allocation) {
        const auto blockIterator = std::find_if(m_blocks.begin(), m_blocks.end(), [&](const auto& currentBlock) {
            if (currentBlock.memory == allocation.memory)
                return true;
            else return false;
        });

        if (blockIterator == m_blocks.end())
            NFError("Could not find video memory block to deallocate from.");

        MemoryBlock& block = *blockIterator;

        const auto allocationIterator = block.allocations.find(allocation.offset);

        if (allocationIterator == block.allocations.end())
            NFError("Could not find video memory allocation to deallocate.");

        block.allocations.erase(allocationIterator);

        // If block is now empty, free it
        if (block.allocations.empty()) {
            vkFreeMemory(m_device, block.memory, nullptr);
            m_blocks.erase(blockIterator);
        }
    }

    void VideoMemoryAllocator::allocate(VkMemoryRequirements memoryRequirements, VkMemoryPropertyFlags memoryPropertyFlags, VideoMemoryAllocation& outAllocation) {
        uint32_t memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, memoryPropertyFlags);

        if (memoryTypeIndex == -1)
            NFError("Could not find suitable memory type.");

        while (true) {
            for (auto& block : m_blocks) {
                if (block.memoryTypeIndex == memoryTypeIndex) {
                    VkDeviceSize lastOffset = 0, lastSize = 0;
                    // Check between existing allocations
                    for (const auto& [currentOffset, currentSize] : block.allocations) {
                        // Check for alignment
                        VkDeviceSize checkOffset = lastOffset + lastSize, remainder = checkOffset % memoryRequirements.alignment;
                        checkOffset = remainder ? checkOffset + (memoryRequirements.alignment - remainder) : checkOffset;
                        // Check for enough space
                        if (currentOffset - lastOffset - lastSize >= memoryRequirements.size) {
                            outAllocation.memory = block.memory;
                            outAllocation.offset = checkOffset;
                            if (block.mappedMemoryPointer)
                                outAllocation.mappedMemoryPointer = reinterpret_cast<char*>(block.mappedMemoryPointer) + outAllocation.offset;
                            block.allocations[checkOffset] = memoryRequirements.size;
                            return;
                        }

                        lastOffset = currentOffset, lastSize = currentSize;
                    }

                    // Check end
                    VkDeviceSize checkOffset = lastOffset + lastSize, remainder = checkOffset % memoryRequirements.alignment;
                    checkOffset = remainder ? checkOffset + (memoryRequirements.alignment - remainder) : checkOffset;
                    if ((block.size - 1) - lastOffset - lastSize >= memoryRequirements.size) {
                        outAllocation.memory = block.memory;
                        outAllocation.offset = checkOffset;
                        if (block.mappedMemoryPointer)
                            outAllocation.mappedMemoryPointer = reinterpret_cast<char*>(block.mappedMemoryPointer) + outAllocation.offset;
                        block.allocations[checkOffset] = memoryRequirements.size;
                        return;
                    }

                    // If not enough space, continue
                    continue;
                }
            }

            // No suitable block found, allocate one
            VkMemoryAllocateInfo newBlockAllocateInfo = { VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO };
            newBlockAllocateInfo.memoryTypeIndex = memoryTypeIndex;
            newBlockAllocateInfo.allocationSize = calculateBlockSize(memoryTypeIndex);
            allocateBlock(newBlockAllocateInfo);

            // If host visible, map
            if (memoryPropertyFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT)
                vkMapMemory(m_device, m_blocks.back().memory, 0, VK_WHOLE_SIZE, 0, &m_blocks.back().mappedMemoryPointer);
        }
    }

    uint32_t VideoMemoryAllocator::findMemoryType(uint32_t typeFilter, VkMemoryPropertyFlags desiredPropertyFlags) {
        VkPhysicalDeviceMemoryProperties memoryProperties = {};
        vkGetPhysicalDeviceMemoryProperties(m_physicalDevice, &memoryProperties);

        for (int i = 0; i < memoryProperties.memoryTypeCount; i++)
            if (typeFilter & (1 << i) && (memoryProperties.memoryTypes[i].propertyFlags & desiredPropertyFlags) == desiredPropertyFlags)
                return i;

        return -1;
    }

    size_t VideoMemoryAllocator::calculateBlockSize(uint32_t memoryTypeIndex) const {
        VkPhysicalDeviceMemoryProperties memoryProperties = {};
        vkGetPhysicalDeviceMemoryProperties(m_physicalDevice, &memoryProperties);
        size_t heapSize = memoryProperties.memoryHeaps[memoryProperties.memoryTypes[memoryTypeIndex].heapIndex].size;

        return heapSize > HEAP_SIZE_LARGE ? DEFAULT_ALLOCATION_AMOUNT : (heapSize / 8);
    }

    void VideoMemoryAllocator::allocateBlock(VkMemoryAllocateInfo& allocateInfo) {
        VkDeviceMemory memoryBlock = nullptr;

        if (vkAllocateMemory(m_device, &allocateInfo, nullptr, &memoryBlock) != VK_SUCCESS)
            NFError("Could not allocate video memory.");

        m_blocks.emplace_back(memoryBlock, allocateInfo.memoryTypeIndex, allocateInfo.allocationSize, nullptr);
    }
}